LILYA_DISCOVER_APP = "LILYA_DEFAULT_APP"
DEFAULT_TEMPLATE_NAME = "default"
APP_PARAMETER = "--app"
HELP_PARAMETER = "--help"
EXCLUDED_DIRECTIVES = ["createproject", "createapp", "createdeployment"]
IGNORE_DIRECTIVES = ["directives"]
DISCOVERY_FILES = ["application.py", "app.py", "main.py"]
DISCOVERY_FUNCTIONS = ["get_application", "get_app"]
TREAT_AS_PROJECT_DIRECTIVE = ["deployment"]
